declare("mdi2", {})


mdi2.grid_size = 20
mdi2.snap_to_grid = true
-- mdi2.active_pdlg

local debug = false

-- ### ######################## ###
-- ### Pseudo-Dialog Definition ###
-- ### ######################## ###

-- PARAMS
-- move
-- resize
-- resizewidth
-- parent
-- cx
-- cy
-- canvasontop

function mdi2.pseudo_dialog(params)
	local old = {}
	assert(type(params) == "table", "Need a table in argument #1")
	assert(iup.IsValid(params.parent) and iup.GetType(params.parent) == "cbox", "PARENT needs to be a CBOX")

	params = params or {}
	params.margin = params.margin or "0x0"
	params.resizewidth = tonumber(params.resizewidth) or params.resize ~= "YES" and 0 or 15

	local pdlg, main, title_bar, contents
	local highlight
	local move_canvas
	local move_layer, move_top, move_bottom, move_left, move_right
	local resize_layer, resize_top, resize_bottom, resize_left, resize_right
	local resize_nw, resize_ne, resize_sw, resize_se

	local resize_bar_width  = tonumber(params.resizewidth)
	local half_resize_width = math.floor(resize_bar_width/2)

	-- never tested with variable widths. may need to re-evaluate placement.  currently places canvases outside of the pdialog
	local marginx, marginy  =  params.margin:match("(%d*)x?(%d*)")
	local adjusted_marginx  = -marginx - resize_bar_width
	local adjusted_marginy  = -marginy - resize_bar_width

	highlight     = iup.stationhighopacityframe{iup.label{title = "", expand = "YES"}, bgcolor = "255 0 0", visible = "NO"}

	move_canvas   = iup.canvas{bgcolor = "101 255 101",     visible = params.move ~= "NO" and "YES" or "NO", border = debug and "YES" or "NO"}

	move_top      = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "101 255 101", visible = "NO",  border = "YES"}
	move_bottom   = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "101 255 101", visible = "NO",  border = "YES"}
	move_left     = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "101 255 101", visible = "NO",  border = "YES"}
	move_right    = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "101 255 101", visible = "NO",  border = "YES"}

	resize_top    = iup.canvas{size = "x"..resize_bar_width,                   expand = "HORIZONTAL", bgcolor = "101 102 255", visible = "NO",  border = debug and "YES" or "NO"}
	resize_bottom = iup.canvas{size = "x"..resize_bar_width,                   expand = "HORIZONTAL", bgcolor = "101 102 255", visible = "NO",  border = debug and "YES" or "NO"}
	resize_left   = iup.canvas{size = resize_bar_width.."x",                   expand = "VERTICAL",   bgcolor = "101 102 255", visible = "NO",  border = debug and "YES" or "NO"}
	resize_right  = iup.canvas{size = resize_bar_width.."x",                   expand = "VERTICAL",   bgcolor = "101 102 255", visible = "NO",  border = debug and "YES" or "NO"}
	resize_nw     = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "0 255 255",   visible = "NO",  border = "YES"}
	resize_ne     = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "0 255 255",   visible = "NO",  border = "YES"}
	resize_sw     = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "0 255 255",   visible = "NO",  border = "YES"}
	resize_se     = iup.canvas{size = resize_bar_width.."x"..resize_bar_width, expand = "NO",         bgcolor = "0 255 255",   visible = "NO",  border = "YES"}


	move_layer = iup.vbox{
		iup.hbox{iup.fill{}, move_top, iup.fill{}},
		iup.vbox{
			iup.fill{},
			iup.hbox{move_left, iup.fill{}, move_right},
			iup.fill{}
		},
		iup.hbox{iup.fill{}, move_bottom, iup.fill{}},
		margin = (adjusted_marginx - resize_bar_width).."x"..(adjusted_marginy - resize_bar_width),
	}


	resize_layer = iup.vbox{
		iup.hbox{
			resize_nw, resize_top, resize_ne
		},
		iup.hbox{
			resize_left, iup.fill{}, resize_right,
		},
		iup.hbox{
			resize_sw, resize_bottom, resize_se
		},
		margin = adjusted_marginx.."x"..adjusted_marginy,
		visible = params.resize ~= "YES" and "NO" or "YES"
	}

	main = iup.zbox{
		params.canvasontop ~= "YES" and move_canvas or params[1],
		params.canvasontop ~= "YES" and params[1] or move_canvas,
		highlight,
		all = "YES",
		alignment = "ACENTER",
	}

	pdlg = iup.pdarootframebg{
		iup.zbox{
			main,
			resize_layer,
			move_layer,
			all = "YES",
		},
		bgcolor = "255 255 255 0 *",
		expand = "NO",
		contents = contents,

		cx = params.cx,
		cy = params.cy,
	}

	-- ### Callbacks ###

	local handles = {
		move_left, move_top, move_right, move_bottom,
		--resize_top, resize_bottom, resize_left, resize_right,
		resize_nw, resize_ne, resize_sw, resize_se
	}

	local doubleclick, dragging, startx, starty, has_moved
	local force_x, force_y = 0, 0 -- How much force has been applied along a certain axis to defeat window reistance TODO apply to touch to prevent accidental movement

	pdlg.mdidragrelease_cb = params.mdidragrelease_cb

	function pdlg:move_to_top()
		local parent = iup.GetParent(pdlg)
		iup.Detach(pdlg)
		iup.Append(parent, pdlg)

		params.parent:map()
	end

	local function get_handle_index(handle)
		if (handle == move_canvas) then return 0 end
		for i, h in ipairs(handles) do
			if (h == handle) then return i end
		end
	end

	local function show_handles()
		for i, h in ipairs(handles) do h.visible = "YES" end
	end

	local function hide_handles(active)
		for i, h in ipairs(handles) do
			if (h ~= active) then h.visible = "NO" end
		end
	end

	function pdlg:highlight(canvas, hide, handles_only, handle_type)
		highlight.visible = handles_only and "YES" or hide and "NO" or "YES"
		highlight.bgcolor = canvas and canvas.bgcolor or highlight.bgcolor
		self.handle_type = handle_type or self.handle_type
		if (handles_only or highlight.visible ~= "YES") then
			hide_handles(handles_only and canvas)
		else
			show_handles()
		end
	end

	function pdlg:move_and_resize(x, y, w, h, nudge, invert)
		-- x|y: can be an absolute position or relative adjustments.
		-- invert changes the operation
		if (not mdi2.active_pdlg) then return end
		if (x) then
			if (nudge) then

				if (pdlg.handle_type == "size") then
					w = tostring(params[1].w)
					h = tostring(params[1].h)
					local abs = math.abs(x)
					w = w + (invert and -abs or abs)

					if (invert) then
						x = x < 0 and -x or 0
					else
						x = x > 0 and 0 or x
					end

				else
					x = invert and -x or x
				end

				x = self.cx + x
			end

			self.cx = x
		end
		if (y) then
			if (nudge) then

				if (pdlg.handle_type == "size") then
					w = tostring(params[1].w)
					h = tostring(params[1].h)
					local abs = math.abs(y)
					h = h + (invert and -abs or abs)

					if (invert) then
						y = y < 0 and -y or 0
					else
						y = y > 0 and 0 or y
					end
				else
					y = invert and -y or y
				end

				y = self.cy + y
			end
			self.cy = y
		end

		if (w and h) then
			params[1].size = w.."x"..h
		end
		iup.Refresh(pdlg)
	end

	local function resize_pdlg_button(self, b, s, x, y, f, properties)
		if (params.button_cb(self, b, s, x, y, f, get_handle_index(self)) == iup.IGNORE) then return end

		local new = {}

		if (s == 1) then -- press
			dragging = true
			startx = x
			starty = y
			pdlg:highlight(self, nil, true, "size")
		else -- release
			if (mdi2.snap_to_grid) then
				for _, prop in ipairs(properties) do
					local r = pdlg[prop] % mdi2.grid_size
					new[prop] = pdlg[prop] - r
				end
				pdlg.cx = new.cx or pdlg.cx
				pdlg.cy = new.cy or pdlg.cy
				params[1].size = (new.w or pdlg.w).."x"..(new.h or pdlg.h)
				pdlg.size = nil
				iup.Refresh(pdlg)
			end

			if (params.mdidragrelease_cb) then pdlg:mdidragrelease_cb(params[1]) end
			dragging = nil
			startx = nil
			starty = nil
			pdlg:highlight(self)
		end
	end

	local function resize_pdlg_motion(self, x, xmod, y, ymod)
		-- xmod and ymod modifies how the delta is applied to the pdlg (values: -1, 0, 1)
		local dx = (x - startx) * xmod
		local dy = (y - starty) * ymod
		local width, height = params[1].w, params[1].h
		width = width + dx
		height = height + dy
		startx = x
		starty = y

		params[1].size = width.."x"..height
		pdlg.size = nil

		-- only adjust the CX and CY values if the adjustment is on the negative side
		if (xmod == -1) then pdlg.cx = pdlg.cx + dx * xmod end
		if (ymod == -1) then pdlg.cy = pdlg.cy + dy * ymod end
		iup.Refresh(pdlg)
	end


	local function move_pdlg_button(self, b, s, x, y, f)
		if (params.button_cb(self, b, s, x, y, f, get_handle_index(self)) == iup.IGNORE) then return end

		if (s == 1) then -- press
			if (b == 8388616) then
				-- Left mouse double click
				doubleclick = true
			elseif (b == 32) then
				-- Right mouse click
			elseif (b == 8 or b == 264) then
				-- Left mouse click or touch press


				if (pdlg ~= mdi2.active_pdlg) then
					if (mdi2.active_pdlg) then
						mdi2.active_pdlg:highlight(self, true)
					end
				end

				dragging = true
				startx = x
				starty = y
				pdlg:highlight(self, nil, true, "move")
			end

		else -- release
			if (b == 32) then
				-- Release right mouse click
			else
				dragging = nil
				startx = nil
				starty = nil

				if (doubleclick) then
					doubleclick = nil
					-- Refresh not needed after this.
					-- Unshading causes listboxes-of-controls to not reshow their contents.  Refresh and Map have no effect on this.
					-- Detach and Append has been proven to redisplay contents, but seems to only be necessary for listboxes.

				else
					if (mdi2.snap_to_grid and has_moved) then
						for axis, leg in pairs({cx = 'w', cy = 'h'}) do
							local r = pdlg[axis] % mdi2.grid_size
							pdlg[axis] = pdlg[axis] - r
						end
						iup.Refresh(pdlg)
					end
					has_moved = nil
					if (params.mdidragrelease_cb) then pdlg:mdidragrelease_cb(params[1]) end

					if (b == nil) then return end -- nil if called after a double-click release, see above.
				end
				if (pdlg ~= mdi2.active_pdlg) then
					mdi2.active_pdlg = pdlg
				end
				pdlg:highlight(self)
			end
		end
	end


	local function move_pdlg_motion(self, x, y, f)
		local dx = x - tonumber(startx)
		local dy = y - tonumber(starty)
		startx = x
		starty = y

		pdlg.cx = tonumber(pdlg.cx) + dx
		pdlg.cy = tonumber(pdlg.cy) + dy
		iup.Refresh(pdlg)
		has_moved = true
	end


	old.button_cb = move_canvas.button_cb
	function move_canvas:button_cb(b, s, x, y, f)
		if (old.button_cb) then old.button_cb(self, b, s, x, y, f) end

		move_pdlg_button(self, b, s, x, y, f)

		if (s == 1) then

		elseif (s == 0) then

			-- Detach and re-Append to move the pdialog to the top of the z-order
			pdlg:move_to_top()
			params.zchange_cb(pdlg)
		end
	end

	old.motion_cb = move_canvas.motion_cb
	function move_canvas:motion_cb(x, y, f)
		if (dragging) then
			if (old.motion_cb) then old.motion_cb(self, x, y, f) end
			move_pdlg_motion(self, x, y, f)
		end
	end

	function move_top:button_cb(b, s, x, y, f)
		move_pdlg_button(self, b, s, x, y, f)
	end

	function move_top:motion_cb(x, y, f)
		if (dragging) then
			move_pdlg_motion(self, x, y, f)
		end
	end

	function move_bottom:button_cb(b, s, x, y, f)
		move_pdlg_button(self, b, s, x, y, f)
	end

	function move_bottom:motion_cb(x, y, f)
		if (dragging) then
			move_pdlg_motion(self, x, y, f)
		end
	end

	function move_left:button_cb(b, s, x, y, f)
		move_pdlg_button(self, b, s, x, y, f)
	end

	function move_left:motion_cb(x, y, f)
		if (dragging) then
			move_pdlg_motion(self, x, y, f)
		end
	end

	function move_right:button_cb(b, s, x, y, f)
		move_pdlg_button(self, b, s, x, y, f)
	end

	function move_right:motion_cb(x, y, f)
		if (dragging) then
			move_pdlg_motion(self, x, y, f)
		end
	end





	old.resize_top_button_cb = resize_top.button_cb
	function resize_top:button_cb(b, s, x, y, f)
		if (old.resize_top_button_cb) then old.resize_top_button_cb(self, b, s, x, y, f) end

		if (s == 1) then -- press
			dragging = true
			startx = x
			starty = y
		else -- release
			resize_pdlg_button(self, b, s, x, y, f, {"cy"})
			dragging = nil
			startx = nil
			starty = nil
		end
	end

	old.resize_top_motion_cb = resize_top.motion_cb
	function resize_top:motion_cb(x, y, f)

		if (dragging) then
			if (old.resize_top_motion_cb) then old.resize_top_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, 0, y, -1)
		end
	end

	old.resize_bottom_button_cb = resize_bottom.button_cb
	function resize_bottom:button_cb(b, s, x, y, f)
		if (old.resize_bottom_button_cb) then old.resize_bottom_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button{self, s, "h"}
	end

	old.resize_bottom_motion_cb = resize_bottom.motion_cb
	function resize_bottom:motion_cb(x, y, f)

		if (dragging) then
			if (old.resize_bottom_motion_cb) then old.resize_bottom_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, 0, y, 1)
		end
	end

	old.resize_left_button_cb = resize_left.button_cb
	function resize_left:button_cb(b, s, x, y, f)
		if (old.resize_left_button_cb) then old.resize_left_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"cx"})
	end

	old.resize_left_motion_cb = resize_left.motion_cb
	function resize_left:motion_cb(x, y, f)

		if (dragging) then
			if (old.resize_left_motion_cb) then old.resize_left_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, -1, y, 0)
		end
	end

	old.resize_right_button_cb = resize_right.button_cb
	function resize_right:button_cb(b, s, x, y, f)
		if (old.resize_right_button_cb) then old.resize_right_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"w"})
	end

	old.resize_right_motion_cb = resize_right.motion_cb
	function resize_right:motion_cb(x, y, f)

		if (dragging) then
			if (old.resize_right_motion_cb) then old.resize_right_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, 1, y, 0)
		end
	end

	old.resize_nw_button_cb = resize_nw.button_cb
	function resize_nw:button_cb(b, s, x, y, f)
		if (old.resize_nw_button_cb) then old.resize_nw_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"cx", "cy"})
	end

	old.resize_nw_motion_cb = resize_nw.motion_cb
	function resize_nw:motion_cb(x, y, f)
		if (dragging and old.resize_nw_motion_cb) then old.resize_nw_motion_cb(self, x, y, f) end

		if (dragging) then
			resize_pdlg_motion(self, x, -1, y, -1)
		end

	end

	old.resize_ne_button_cb = resize_ne.button_cb
	function resize_ne:button_cb(b, s, x, y, f)
		if (old.resize_ne_button_cb) then old.resize_ne_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"cy", "w"})
	end

	old.resize_ne_motion_cb = resize_ne.motion_cb
	function resize_ne:motion_cb(x, y, f)
		if (dragging and old.resize_ne_motion_cb) then old.resize_ne_motion_cb(self, x, y, f) end

		if (dragging) then
			if (old.resize_top_motion_cb) then old.resize_top_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, 1, y, -1)
		end
	end

	old.resize_sw_button_cb = resize_sw.button_cb
	function resize_sw:button_cb(b, s, x, y, f)
		if (old.resize_sw_button_cb) then old.resize_sw_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"cx", "h"})
	end

	old.resize_sw_motion_cb = resize_sw.motion_cb
	function resize_sw:motion_cb(x, y, f)
		if (dragging and old.resize_sw_motion_cb) then old.resize_sw_motion_cb(self, x, y, f) end

		if (dragging) then
			if (old.resize_top_motion_cb) then old.resize_top_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, -1, y, 1)
		end
	end

	old.resize_se_button_cb = resize_se.button_cb
	function resize_se:button_cb(b, s, x, y, f)
		if (old.resize_se_button_cb) then old.resize_se_button_cb(self, b, s, x, y, f) end

		resize_pdlg_button(self, b, s, x, y, f, {"w", "h"})
	end

	old.resize_se_motion_cb = resize_se.motion_cb
	function resize_se:motion_cb(x, y, f)
		if (dragging and old.resize_se_motion_cb) then old.resize_se_motion_cb(self, x, y, f) end

		if (dragging) then
			if (old.resize_top_motion_cb) then old.resize_top_motion_cb(self, x, y, f) end
			resize_pdlg_motion(self, x, 1, y, 1)
		end
	end
	iup.Append(iup.GetNextChild(params.parent), pdlg)
	iup.Map(params.parent)

	if (params.cx == "ACENTER") then
		pdlg.cx = math.floor(DroidButtons.xres / 2) - math.floor(pdlg.w / 2)
	end
	if (params.cy == "ACENTER") then
		pdlg.cy = math.floor(DroidButtons.yres / 2) - math.floor(pdlg.h / 2)
	end
	iup.Refresh(pdlg)
	return pdlg
end
